package chatoyant

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Chatoyant_runtime.EffectSource

Effect signatures used to keep pure SDK logic independent from runtime IO.

Native OCaml, Node.js through Melange, browsers, and tests should provide modules matching these signatures. Core/provider modules should depend on signatures through functors instead of reaching directly for global primitives.

Sourcemodule type ENV = sig ... end
Sourcemodule type CLOCK = sig ... end
Sourcemodule type HTTP = sig ... end
Sourcemodule type WEBSOCKET = sig ... end