package chatoyant

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Json_schema.ValidatorSource

Draft 2020-12 validation engine.

The validator implements assertion keywords and tracks evaluated object properties/array items for the unevaluated vocabulary. Format is treated as an annotation by default, matching draft 2020-12 unless the caller opts into format assertions.

Sourcetype options = {
  1. format_assertion : bool;
  2. resources : Chatoyant_schema__.Json_schema_resolver.resource list;
}
Sourceval default_options : options
Sourceval validate : ?options:options -> Chatoyant_schema__.Json_schema_ast.t -> Chatoyant_runtime.Json.t -> Chatoyant_schema__.Json_schema_output.t
Sourceval validate_json : ?options:options -> Chatoyant_runtime.Json.t -> Chatoyant_runtime.Json.t -> (Chatoyant_schema__.Json_schema_output.t, Chatoyant_schema__.Json_schema_ast.parse_error) result